% Cell Tracking Challenge 2015
% Dr. Omer Demirel - 02/2015
function [] = CellTracker(input_seq,output_seq,thresholdfactor,kernelWidth,imextendedParam)

maxDist = 50;
maxNN = 5;
maxCellGrowth = 2.0;

thresholdfactor = str2double(thresholdfactor);
kernelWidth = str2num(kernelWidth);
imextendedParam = str2double(imextendedParam);

cd(input_seq);

% dataNum = ['0' int2str(ii)];

allTiffFiles = dir('*.tif');
numFiles = length(allTiffFiles);
range=1:numFiles;

%% Parameters
maxObj=400;
tracks = zeros(numFiles,maxObj,3)-2;
numObjs = zeros(1,numFiles);
for thisFrame=range
    I = read3D(allTiffFiles(thisFrame).name);
    I = simple_preprocessor(I,kernelWidth);
    S = simple_advWatershed(I,thresholdfactor,imextendedParam);
    S = bwlabel(S,8);
    cc = bwconncomp(S,8);
    thisStats = regionprops(cc,S,'Area','Centroid');
    thisObj = cat(1, thisStats.Centroid);
    %% update # objects
    numObjs(thisFrame)=length(thisObj);
    %% check next frame
    if thisFrame~=numFiles
        nextFrame = thisFrame+1;
        I_next = read3D(allTiffFiles(nextFrame).name);
        I_next = simple_preprocessor(I_next,kernelWidth);
        S_next = simple_advWatershed(I_next,thresholdfactor,imextendedParam);
        S_next = bwlabel(S_next,8);
        cc_next = bwconncomp(S_next,8);
        
        nextStats = regionprops(cc_next,S_next,'Area','Centroid');
        nextObj = cat(1, nextStats.Centroid);
        
        %% compute all-to-all nearest neighbors
        [NNs,allDist] = getNearestObjIds(thisObj,nextObj,maxNN,maxDist);
        %% for each object in thisFrame, compute possible candidates
        nextId=maxObj;
        for i=1:length(thisObj)
            nextIds = NNs(i,:);
            nextIds(nextIds==0)=[];
            %% maximum distance filtering
            nextDists = allDist(i,nextIds);
            nextIds(nextDists>maxDist)=[];
            candIds = nextIds;
            thisArea = cat(1, thisStats(i).Area);
            nextAreas = cat(1, nextStats(candIds).Area);
            %% build simple tracklets
            if isempty(candIds)
                %                 display('no neighbors.. Disappearing object?')
                tracks(thisFrame,i,:)=[i 0 -1];
                continue
            elseif length(candIds)==1 && sum(nextAreas)<maxCellGrowth*thisArea
                %                 display('single neighbor... probably same object')
                tracks(thisFrame,i,:)=[i candIds(1) -1];
                continue
            elseif length(candIds)==1 && sum(nextAreas)>maxCellGrowth*thisArea
                %                 sum(nextAreas)/thisArea
                tracks(thisFrame,i,:)=[i 0 -1];
                %plot2Frames(i,[-1 -1],thisStats,nextStats,S,S_next,thisFrame,'NO Self-tracking - Disappearing!')
                continue
            elseif length(candIds)==1
                tracks(thisFrame,i,:)=[i candIds(1) -1];
                continue
            else
                %                 display('multiple neighbors... further analysis')
            end
            candDist = allDist(i,candIds);
            if ~isempty(find(candDist < 10))
                %                 display('Probably same object')
                tracks(thisFrame,i,:)= [i candIds(1) -1];
                %plot2Frames(i,candIds(1),thisStats,nextStats,S,S_next,thisFrame,'Self tracking')
                continue
            end
            %% Cell division?
            % we assume that we have >=2 candidates, find all the centroids
            % of all possible candidate pairs. choose the ones that
            % produced the nearest average centroid position
            nextCentroid = cat(1, nextStats(candIds).Centroid);
            thisCentroid = cat(1, thisStats(i).Centroid);
            nearestPair = [];
            minDist = Inf;
            for j=1:length(nextCentroid)-1
                for k=j+1:length(nextCentroid)
                    xy = (nextCentroid(j,:)+nextCentroid(k,:))./2;
                    currDist = pdist2(thisCentroid,xy);
                    if minDist>currDist && currDist < maxDist
                        minDist = currDist;
                        nearestPair = [i candIds(j) candIds(k)];
                    end
                end
            end
            cleanNearestPair = nearestPair(2:3);
            nextAreas = cat(1, nextStats(cleanNearestPair).Area);
            if sum(nextAreas)>maxCellGrowth*thisArea
                %                 display('Nope...')
                nextCenters = cat(1, nextStats(cleanNearestPair).Centroid);
                dupThisCentroid = [thisCentroid;thisCentroid];
                currDist = pdist2(thisCentroid,nextCenters);
                [m,ind] = min(currDist);
                tracks(thisFrame,i,:)=[i cleanNearestPair(ind) -1];
                continue
            else
                tracks(thisFrame,i,:) = nearestPair;
                continue
            end
            
        end
    else
        for i=1:length(thisObj)
            tracks(thisFrame,i,:)=[i 0 -1];
        end
    end
    %% track check: a cell cannot have two mothers
    a = find(tracks(thisFrame,:,2)>0);
    u = unique(a); n = hist(a,u); idx = u(n>1);
    for dup=1:length(idx)
        %         display('Two parents!')
        matches = find(a==idx(dup));
        tracks(thisFrame,matches(2:length(matches)),2)=0;
        tracks(thisFrame,matches(2:length(matches)),3)=-1;
    end
    %% Write segmentation results
    S = uint16(S);
    if thisFrame>100
        filename = int2str(thisFrame-1);
    elseif thisFrame>10
        filename = ['0' int2str(thisFrame-1)];
    else
        filename = ['00' int2str(thisFrame-1)];
    end
    %     cd(['../' dataNum '_RES']);
    cd('../../SW');
    cd(output_seq);
    imwrite(S,['mask' filename '.tif']);
    cd('../../SW');
    cd(input_seq);
    
end
% end
%     prevTracks = tracks;
% convert all Ids to unique ones
nextAvailId = numObjs(1);
for thisFrame=1:numFiles-1
    nextFrame=thisFrame+1;
    
    for nextObj=1:numObjs(nextFrame)
        found = find(tracks(nextFrame,nextObj,1)==tracks(thisFrame,:,2));
        found2 = find(tracks(nextFrame,nextObj,1)==tracks(thisFrame,:,3));
        if tracks(thisFrame,found,3)==-1 % self-tracking OR disappearing
            if ~isempty(found) % self-tracking
                if length(found)>1
                    f = found(length(found));
                else
                    f = found;
                end
                tracks(nextFrame,nextObj,1)=tracks(thisFrame,f,1);
            else % disappearing
                nextAvailId=nextAvailId+1;
            end
        elseif ~isempty(found) || ~isempty(found2)
            nextAvailId=nextAvailId+1;
            tracks(nextFrame,nextObj,1)=nextAvailId;
        else % appearing
            nextAvailId=nextAvailId+1;
            tracks(nextFrame,nextObj,1)=nextAvailId;
            %             display(['Something is wrong!! Frame:' int2str(nextFrame) ' #' int2str(nextObj)])
        end
    end
end

%% update TIFF labels with correct ones

for thisFrame=1:numFiles
    if thisFrame>100
        filename = int2str(thisFrame-1);
    elseif thisFrame>10
        filename = ['0' int2str(thisFrame-1)];
    else
        filename = ['00' int2str(thisFrame-1)];
    end
    %     cd(['../' dataNum '_RES']);
    cd('../../SW');
    cd(output_seq);
    
    S = imread(['mask' filename '.tif']);
    S = int16(S);
    for i=1:numObjs(thisFrame)
        if i==tracks(thisFrame,i,1)
            S(S==i)=tracks(thisFrame,i,1);
        else
            S(S==i)=-tracks(thisFrame,i,1);
        end
    end
    yunik = unique(S);
    ind = find(yunik<0);
    changes = yunik(ind);
    for i=1:length(changes)
        S(S==changes(i))=abs(changes(i));
    end
    S = uint16(S);
    if find(S<0)
        display('WRONG')
    end
    
    imwrite(S,['mask' filename '.tif']);
    cd('../../SW');
    cd(input_seq);
    %     cd(['../' dataNum]);
end

% C(1:5,1:3)=tracks(1,1:5,1: 3)
% C(1:5,1:3)=tracks(2,1:5,1: 3)
% C(1:5,1:3)=tracks(3,1:5,1: 3)
% C(1:5,1:3)=tracks(4,1:5,1: 3)

trackFile=[];
% find all tracks including subtracks.
mapParents = containers.Map('KeyType','int32','ValueType','int32');
for thisFrame=1:numFiles
    for i=1:numObjs(thisFrame)
        thisObj = tracks(thisFrame,i,1);
        nextFrame = thisFrame+1;
        mapParents(thisObj) = -1;
        
        while nextFrame<=numFiles
            allNextObj = tracks(nextFrame,:,1);
            found = find(allNextObj==thisObj);
            if isempty(found)
                break
            end
            nextFrame = nextFrame+1;
        end
        endFrame = nextFrame-1;
        trackEntry = [thisObj thisFrame-1 endFrame-1 mapParents(thisObj)];
        trackFile = [trackFile;trackEntry];
    end
end
trackFile = sortrows(trackFile,1);

% remove subtracks
uniqueIds = unique(trackFile(:,1));
for id = 1:length(uniqueIds)
    % find all rows that start with this id
    found = find(trackFile(:,1)==id);
    shortlistRows = trackFile(found,:);
    minStart = min(shortlistRows(:,2));
    %     if
    maxEnd   = max(shortlistRows(:,3));
    trackFile(found,2)=minStart;
    trackFile(found,3)=maxEnd;
end
trackFile = unique(trackFile,'rows');

% find all parents
for thisFrame=numFiles:-1:2
    for i=1:numObjs(thisFrame)
        thisObj=tracks(thisFrame,i,1);
        prevFrame=thisFrame-1;
        found = find(tracks(prevFrame,:,1)==thisObj);
        while ~isempty(found)
            if prevFrame==1
                break
            end
            prevFrame=prevFrame-1;
            found = find(tracks(prevFrame,:,1)==thisObj);
        end
        foundAt2 = find(tracks(prevFrame,:,2)==i);
        foundAt3 = find(tracks(prevFrame,:,3)==i);
        if ~isempty(foundAt2)
            parent = tracks(prevFrame,foundAt2,1);
        elseif ~isempty(foundAt3)
            parent = tracks(prevFrame,foundAt3,1);
        else
            parent = 0;
        end
        %             display([ int2str(thisObj) ' in Frame ' int2str(thisFrame) ' has parent ' int2str(parent)])
        if length(parent)>1
            mapParents(thisObj)=parent(1);
        else
            mapParents(thisObj)=parent;
        end
    end
end
for i=1:numObjs(1)
    thisObj=tracks(1,i,1);
    mapParents(thisObj)=0;
end
% add correct parents to the trackfile
for entry = 1:length(trackFile)
    trackFile(entry,:) = [trackFile(entry,1:3) mapParents(entry)];
end
% cd(['../' dataNum '_RES']);
cd('../../SW');
cd(output_seq);
dlmwrite('res_track.txt',trackFile, ' ');
% cd(['../' dataNum]);
cd('../../SW');
cd(input_seq);

% display(['Dataset ' dataNum ' is done!'])





end
